/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             Ext2.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ntifs.h"
#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2LoadSuper)
#pragma alloc_text(PAGE, Ext2SaveSuper)

#pragma alloc_text(PAGE, Ext2LoadGroup)
#pragma alloc_text(PAGE, Ext2SaveGroup)

#pragma alloc_text(PAGE, Ext2GetInodeLba)
#pragma alloc_text(PAGE, Ext2LoadInode)
#pragma alloc_text(PAGE, Ext2SaveInode)

#pragma alloc_text(PAGE, Ext2LoadBlock)
#pragma alloc_text(PAGE, Ext2SaveBlock)

#pragma alloc_text(PAGE, Ext2SaveBuffer)

#pragma alloc_text(PAGE, Ext2GetBlock)
#pragma alloc_text(PAGE, Ext2BlockMap)

#pragma alloc_text(PAGE, Ext2BuildBDL)

#pragma alloc_text(PAGE, Ext2NewBlock)
#pragma alloc_text(PAGE, Ext2FreeBlock)

#pragma alloc_text(PAGE, Ext2ExpandBlock)
#pragma alloc_text(PAGE, Ext2ExpandInode)

#pragma alloc_text(PAGE, Ext2NewInode)
#pragma alloc_text(PAGE, Ext2FreeInode)

#pragma alloc_text(PAGE, Ext2AddEntry)
#pragma alloc_text(PAGE, Ext2RemoveEntry)

#pragma alloc_text(PAGE, Ext2TruncateBlock)
#pragma alloc_text(PAGE, Ext2TruncateInode)

#pragma alloc_text(PAGE, Ext2AddMcbEntry)
#pragma alloc_text(PAGE, Ext2RemoveMcbEntry)
#pragma alloc_text(PAGE, Ext2LookupMcbEntry)

#endif

/* FUNCTIONS ***************************************************************/

PEXT2_SUPER_BLOCK
Ext2LoadSuper(IN PEXT2_VCB      Vcb,
              IN BOOLEAN        bVerify )
{
    NTSTATUS          Status;
    PEXT2_SUPER_BLOCK Ext2Sb = NULL;
    
    Ext2Sb = (PEXT2_SUPER_BLOCK) ExAllocatePool( PagedPool, 
                                                 SUPER_BLOCK_SIZE);
    if (!Ext2Sb)
    {
        return NULL;
    }

    Status = Ext2ReadDisk( Vcb,
                           (ULONGLONG) SUPER_BLOCK_OFFSET,
                           SUPER_BLOCK_SIZE,
                           (PVOID) Ext2Sb,
                           bVerify );

    if (!NT_SUCCESS(Status))
    {
            Ext2Print((DBG_ERROR, "Ext2ReadDisk: Read Block Device error.\n"));

            ExFreePool(Ext2Sb);
            return NULL;
    }

    return Ext2Sb;
}


BOOLEAN
Ext2SaveSuper(  IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb )
{
    LONGLONG    Offset;
    BOOLEAN     bRet;

    Offset = (LONGLONG) SUPER_BLOCK_OFFSET;

    bRet = Ext2SaveBuffer( IrpContext,
                           Vcb,
                           Offset,
                           SUPER_BLOCK_SIZE,
                           Vcb->ext2_super_block );

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK))
    {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}


BOOLEAN
Ext2LoadGroup(IN PEXT2_VCB Vcb)
{
    ULONG       Size;
    PVOID       Buffer;
    LONGLONG    Lba;
    NTSTATUS    Status;

    PEXT2_SUPER_BLOCK sb;

    sb = Vcb->ext2_super_block;

    Vcb->BlockSize  = EXT2_MIN_BLOCK << sb->s_log_block_size;
    Vcb->SectorBits = Ext2Log2(SECTOR_SIZE);
    ASSERT(BLOCK_BITS == Ext2Log2(BLOCK_SIZE));

    Vcb->ext2_groups = (sb->s_blocks_count - sb->s_first_data_block +
        sb->s_blocks_per_group - 1) / sb->s_blocks_per_group;

    Size = sizeof(EXT2_GROUP_DESC) * Vcb->ext2_groups;

    if (Vcb->BlockSize == EXT2_MIN_BLOCK)
    {
        Lba = (LONGLONG)2 * Vcb->BlockSize;
    }

    if (Vcb->BlockSize > EXT2_MIN_BLOCK)
    {
        Lba = (LONGLONG) (Vcb->BlockSize);
    }

    if (Lba == 0)
    {
        return FALSE;
    }

    Buffer = ExAllocatePool(PagedPool, Size);
    if (!Buffer)
    {
        Ext2Print((DBG_ERROR, "Ext2LoadSuper: no enough memory.\n"));
        return FALSE;
    }

    Ext2Print((DBG_INFO, "Ext2LoadGroup: Lba=%I64xh Size=%xh\n", Lba, Size));

    Status = Ext2ReadDisk(  Vcb,
                            Lba,
                            Size,
                            Buffer,
                            FALSE );

    if (!NT_SUCCESS(Status))
    {
        ExFreePool(Buffer);
        Buffer = NULL;

        return FALSE;
    }

    Vcb->ext2_group_desc = (PEXT2_GROUP_DESC) Buffer;

    return TRUE;
}


BOOLEAN
Ext2SaveGroup(  IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb )
{
    LONGLONG    Offset;
    ULONG       Len;
    BOOLEAN     bRet;

    if (Vcb->BlockSize == EXT2_MIN_BLOCK) {

        Offset = (LONGLONG) (2 * Vcb->BlockSize);

    } else {

        Offset = (LONGLONG) (Vcb->BlockSize);
    }

    Len = (ULONG)(sizeof(struct ext2_group_desc) * Vcb->ext2_groups);

    bRet = Ext2SaveBuffer( IrpContext, Vcb, Offset,
                           Len, Vcb->ext2_group_desc );

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK))
    {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}

BOOLEAN
Ext2GetInodeLba (IN PEXT2_VCB   Vcb,
         IN  ULONG  inode,
         OUT PLONGLONG offset)
{
    LONGLONG loc;

    if (inode < 1 || inode > INODES_COUNT)
    {
        Ext2Print((DBG_ERROR, "Ext2GetInodeLba: Inode value %xh is invalid.\n",inode));
        *offset = 0;
        return FALSE;
    }

    loc = (Vcb->ext2_group_desc[(inode - 1) / INODES_PER_GROUP ].bg_inode_table);
    loc = loc * Vcb->BlockSize;
    loc = loc + ((inode - 1) % INODES_PER_GROUP) * sizeof(EXT2_INODE);

    *offset = loc;

    return TRUE;
}

BOOLEAN
Ext2LoadInode (IN PEXT2_VCB Vcb,
           IN ULONG     inode,
           IN PEXT2_INODE ext2_inode)
{
    IO_STATUS_BLOCK     IoStatus;
    LONGLONG            Offset; 

    if (!Ext2GetInodeLba(Vcb, inode, &Offset))
    {
        Ext2Print((DBG_ERROR, "Ext2LoadInode: error get inode(%xh)'s addr.\n", inode));
        return FALSE;
    }

    if (!Ext2CopyRead(
            Vcb->StreamObj,
            (PLARGE_INTEGER)&Offset,
            sizeof(EXT2_INODE),
            PIN_WAIT,
            (PVOID)ext2_inode,
            &IoStatus ));

    if (!NT_SUCCESS(IoStatus.Status))
    {
        return FALSE;
    }

    return TRUE;
}

/*
BOOLEAN
Ext2SaveInode (IN PEXT2_VCB Vcb,
           IN ULONG inode,
           IN struct ext2_inode *ext2_inode)
{
    ULONG       lba;
    ULONG       offset;
    NTSTATUS    Status;

    if (!Ext2GetInodeLba(Vcb, inode, &lba, &offset))
    {
        Ext2Print((DBG_ERROR, "Ext2LoadInode: error get inode(%xh)'s addr.\n", inode));
        return FALSE;
    }

    Status = Ext2WriteDisk(Vcb->TargetDeviceObject,
        lba,
        offset,
        sizeof(EXT2_INODE),
        (PVOID)ext2_inode);

    if (!NT_SUCCESS(Status))
    {
        return FALSE;
    }

    return TRUE;
}
*/

BOOLEAN
Ext2SaveInode ( IN PEXT2_IRP_CONTEXT IrpContext,
                IN PEXT2_VCB Vcb,
                IN ULONG Inode,
                IN PEXT2_INODE Ext2Inode)
{
    LONGLONG        Offset = 0;
    LARGE_INTEGER   CurrentTime;
    BOOLEAN         bRet;

    KeQuerySystemTime(&CurrentTime);
    Ext2Inode->i_mtime = Ext2Inode->i_atime = 
                      (ULONG)(Ext2InodeTime(CurrentTime));

    Ext2Print((DBG_INFO, "Ext2SaveInode: Saving Inode %xh: Mode=%xh Size=%xh\n",
                         Inode, Ext2Inode->i_mode, Ext2Inode->i_size));

    if (!Ext2GetInodeLba(Vcb, Inode, &Offset))
    {
        Ext2Print((DBG_ERROR, "Ext2SaveInode: error get inode(%xh)'s addr.\n", Inode));
        return FALSE;
    }

    bRet = Ext2SaveBuffer(IrpContext, Vcb, Offset, sizeof(EXT2_INODE), Ext2Inode);

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK))
    {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}


BOOLEAN
Ext2LoadBlock (IN PEXT2_VCB Vcb,
           IN ULONG     dwBlk,
           IN PVOID     Buffer )
{
    IO_STATUS_BLOCK     IoStatus;
    LONGLONG            Offset; 

    Offset = (LONGLONG) dwBlk;
    Offset = Offset * Vcb->BlockSize;

    if (!Ext2CopyRead(
            Vcb->StreamObj,
            (PLARGE_INTEGER)&Offset,
            Vcb->BlockSize,
            PIN_WAIT,
            Buffer,
            &IoStatus ));

    if (!NT_SUCCESS(IoStatus.Status))
    {
        return FALSE;
    }

    return TRUE;
}


BOOLEAN
Ext2SaveBlock ( IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb,
                IN ULONG                dwBlk,
                IN PVOID                Buf )
{
    LONGLONG Offset;
    BOOLEAN  bRet;

    Offset = (LONGLONG) dwBlk;
    Offset = Offset * Vcb->BlockSize;

    bRet = Ext2SaveBuffer(IrpContext, Vcb, Offset, Vcb->BlockSize, Buf);

    if (IsFlagOn(Vcb->Flags, VCB_FLOPPY_DISK))
    {
        Ext2StartFloppyFlushDpc(Vcb, NULL, NULL);
    }

    return bRet;
}


BOOLEAN
Ext2SaveBuffer( IN PEXT2_IRP_CONTEXT    IrpContext,
                IN PEXT2_VCB            Vcb,
                IN LONGLONG             Offset,
                IN ULONG                Size,
                IN PVOID                Buf )
{
    PBCB        Bcb;
    PVOID       Buffer;

    if( !CcPinRead( Vcb->StreamObj,
                    (PLARGE_INTEGER) (&Offset),
                    Size,
                    PIN_WAIT,
                    &Bcb,
                    &Buffer ))
    {
        Ext2Print((DBG_ERROR, "Ext2SaveBuffer: PinReading error ...\n"));
        return FALSE;
    }


    Ext2Print((DBG_INFO, "Ext2SaveBuffer: Off=%I64xh Len=%xh Bcb=%xh\n",
                         Offset, Size, (ULONG)Bcb));

    RtlCopyMemory(Buffer, Buf, Size);
    CcSetDirtyPinnedData(Bcb, NULL );

    Ext2RepinBcb(IrpContext, Bcb);

    CcUnpinData(Bcb);

    SetFlag(Vcb->StreamObj->Flags, FO_FILE_MODIFIED);

    Ext2AddMcbEntry(Vcb, Offset, (LONGLONG)Size);

    return TRUE;
}

ULONG Ext2GetBlock(IN PEXT2_VCB Vcb,
           ULONG dwContent,
           ULONG Index,
           int   layer )
{
    ULONG       *pData = NULL;
    ULONG       i = 0, j = 0, temp = 1;
    ULONG       dwBlk = 0;

    if (layer == 0)
    {
        dwBlk = dwContent;
    }
    else if (layer <= 3)
    {
        pData = (ULONG *)ExAllocatePool(PagedPool,
                    Vcb->BlockSize);
        if (!pData)
        {
            Ext2Print((DBG_ERROR, "Ext2GetBlock: no enough memory.\n"));
            return dwBlk;
        }

        if (!Ext2LoadBlock(Vcb, dwContent, pData))
        {
            ExFreePool(pData);
            return 0;
        }

        temp = 1 << ((BLOCK_BITS - 2) * (layer - 1));
        
        i = Index / temp;
        j = Index % temp;

        dwBlk = pData[i];

        ExFreePool(pData);

        dwBlk = Ext2GetBlock(Vcb, dwBlk, j, layer - 1);
    }
    
    return dwBlk;
}

ULONG Ext2BlockMap(IN PEXT2_VCB Vcb,
           IN PEXT2_INODE ext2_inode,
           IN ULONG Index)
{
    ULONG   dwSizes[EXT2_BLOCK_TYPES];
    int     i;
    ULONG   dwBlk;
    ULONG   Totalblocks;

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        dwSizes[i] = Vcb->dwData[i];
    }

    Totalblocks = ((ext2_inode->i_blocks) >> (BLOCK_BITS - SECTOR_BITS));

    if (Index >= Ext2DataBlocks(Vcb, Totalblocks))
    {
        Ext2Print((DBG_ERROR, "Ext2BlockMap: error input paramters.\n"));

        Ext2BreakPoint();

        return 0;
    }

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        if (Index < dwSizes[i])
        {
            dwBlk = ext2_inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)];
#if DBG
            {   
                ULONG dwRet = Ext2GetBlock(Vcb, dwBlk, Index , i);

                Ext2Print((DBG_INFO, "Ext2BlockMap: i=%xh index=%xh dwBlk=%xh (%xh)\n", i, Index, dwRet, dwBlk));

                return dwRet;
            }
#else
            return Ext2GetBlock(Vcb, dwBlk, Index , i);
#endif
        }

        Index -= dwSizes[i];
    }

    return 0;
}


ULONG Ext2BuildBDL( IN PEXT2_IRP_CONTEXT    IrpContext,
                    IN PEXT2_VCB Vcb,
                    IN PEXT2_INODE ext2_inode,
                    IN ULONG Offset, 
                    IN ULONG Size, 
                    OUT PEXT2_BDL *ext2_bdl )
{
    ULONG    nBeg, nEnd, nBlocks;
    ULONG    dwBlk, i;
    ULONG    dwBytes = 0;
    LONGLONG Lba;
    LONGLONG AllocSize;
    ULONG    Totalblocks;

    PEXT2_BDL   ext2bdl;

    *ext2_bdl = NULL;

    
    Totalblocks = (ext2_inode->i_blocks >> (BLOCK_BITS - SECTOR_BITS));
    AllocSize = Ext2DataBlocks(Vcb, Totalblocks);
    AllocSize = (AllocSize << BLOCK_BITS);

    if ((LONGLONG)Offset >= AllocSize)
    {
        Ext2Print((DBG_ERROR, "Ext2BuildBDL: beyond the file range.\n"));
        return 0;
    }

    if ((LONGLONG)(Offset + Size) > AllocSize)
    {
        Size = (ULONG)(AllocSize - Offset);
    }

    nBeg = (Offset >> BLOCK_BITS);
    nEnd = ((Size + Offset + Vcb->BlockSize - 1) >> BLOCK_BITS);

    nBlocks = 0;

    if ((nEnd - nBeg) > 0)
    {
        ext2bdl = ExAllocatePool(PagedPool, sizeof(EXT2_BDL) * (nEnd - nBeg));

        if (ext2bdl)
        {

            RtlZeroMemory(ext2bdl, sizeof(EXT2_BDL) * (nEnd - nBeg));
            
            for (i = nBeg; i < nEnd; i++)
            {
                dwBlk = Ext2BlockMap(Vcb, ext2_inode, i);

                if (dwBlk > 0)
                {
                    Lba = (LONGLONG) dwBlk;
                    Lba = Lba * Vcb->BlockSize;
                
                    if (nBeg == nEnd - 1) // ie. (nBeg == nEnd - 1)
                    {
                        dwBytes = Size;
                        ext2bdl[nBlocks].Lba = Lba + (LONGLONG)(Offset % (Vcb->BlockSize));
                        ext2bdl[nBlocks].Length = dwBytes;
                        ext2bdl[nBlocks].Offset = 0;

                        nBlocks++;
                    }
                    else
                    {
                        if (i == nBeg)
                        {
                            dwBytes = Vcb->BlockSize - (Offset % (Vcb->BlockSize));
                            ext2bdl[nBlocks].Lba = Lba + (LONGLONG)(Offset % (Vcb->BlockSize));
                            ext2bdl[nBlocks].Length = dwBytes;
                            ext2bdl[nBlocks].Offset = 0;

                            nBlocks++;
                        }
                        else if (i == nEnd - 1)
                        {
                            if (ext2bdl[nBlocks - 1].Lba + ext2bdl[nBlocks - 1].Length == Lba)
                            {
                                ext2bdl[nBlocks - 1].Length += Size - dwBytes;
                            }
                            else
                            {
                                ext2bdl[nBlocks].Lba = Lba;
                                ext2bdl[nBlocks].Length = Size - dwBytes;
                                ext2bdl[nBlocks].Offset = dwBytes;
                                nBlocks++;
                            }

                            dwBytes = Size;

                        }
                        else
                        {
                            if (ext2bdl[nBlocks - 1].Lba + ext2bdl[nBlocks - 1].Length == Lba)
                            {
                                ext2bdl[nBlocks - 1].Length += Vcb->BlockSize;
                            }
                            else
                            {
                                ext2bdl[nBlocks].Lba = Lba;
                                ext2bdl[nBlocks].Length = Vcb->BlockSize;
                                ext2bdl[nBlocks].Offset = dwBytes;
                                nBlocks++;
                            }

                            dwBytes +=  Vcb->BlockSize;
                        }
                    }
                }
                else
                {
                    break;
                }
            }

            *ext2_bdl = ext2bdl;
            return nBlocks;
        }
    }

    // Error
    return 0;
}

BOOLEAN Ext2NewBlock(   PEXT2_IRP_CONTEXT IrpContext,
                        PEXT2_VCB Vcb,
                        ULONG     GroupHint,
                        ULONG     BlockHint,  
                        PULONG    dwRet )
{
    RTL_BITMAP      BlockBitmap;
    LARGE_INTEGER   Offset;
    ULONG           Length;

    PBCB            BitmapBcb;
    PVOID           BitmapCache;

    ULONG           Group = 0, dwBlk, dwHint = 0;

    *dwRet = 0;
    dwBlk = 0XFFFFFFFF;

    if (GroupHint > Vcb->ext2_groups)
        GroupHint = Vcb->ext2_groups - 1;

    if (BlockHint != 0)
    {
        GroupHint = (BlockHint - EXT2_FIRST_DATA_BLOCK) / BLOCKS_PER_GROUP;
        dwHint = (BlockHint - EXT2_FIRST_DATA_BLOCK) % BLOCKS_PER_GROUP;
    }
  
ScanBitmap:
  
    // Perform Prefered Group
    if (Vcb->ext2_group_desc[GroupHint].bg_free_blocks_count)
    {
        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * 
                          Vcb->ext2_group_desc[GroupHint].bg_block_bitmap;

        if (GroupHint == Vcb->ext2_groups - 1)
        {
            Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

            /* s_blocks_count is integer multiple of s_blocks_per_group */
            if (Length == 0)
            {
                Length = BLOCKS_PER_GROUP;
            }
        }
        else
        {
            Length = BLOCKS_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) )
        {
            Ext2Print((DBG_ERROR, "Ext2NewBlock: PinReading error ...\n"));
            return FALSE;
        }

        RtlInitializeBitMap( &BlockBitmap,
                             BitmapCache,
                             Length );

        Group = GroupHint;

        if (RtlCheckBit(&BlockBitmap, dwHint) == 0)
        {
            dwBlk = dwHint;
        }
        else
        {
            dwBlk = RtlFindClearBits(&BlockBitmap, 1, dwHint);
        }

        // We could not get new block in the prefered group.
        if (dwBlk == 0xFFFFFFFF)
        {
            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
        }
    }

    if (dwBlk == 0xFFFFFFFF)
    {
        for(Group = 0; Group < Vcb->ext2_groups; Group++)
        if (Vcb->ext2_group_desc[Group].bg_free_blocks_count)
        {

            if (Group == GroupHint)
                continue;

            Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
            Offset.QuadPart = Offset.QuadPart * Vcb->ext2_group_desc[Group].bg_block_bitmap;

            if (Vcb->ext2_groups == 1)
            {
                Length = TOTAL_BLOCKS;
            }
            else
            {
                if (Group == Vcb->ext2_groups - 1)
                {
                    Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

                    /* s_blocks_count is integer multiple of s_blocks_per_group */
                    if (Length == 0)
                    {
                        Length = BLOCKS_PER_GROUP;
                    }
                }
                else
                {
                    Length = BLOCKS_PER_GROUP;
                }
            }

            if (!CcPinRead( Vcb->StreamObj,
                            &Offset,
                            Vcb->BlockSize,
                            PIN_WAIT,
                            &BitmapBcb,
                            &BitmapCache ) )
            {
                Ext2Print((DBG_ERROR, "Ext2NewBlock: PinReading error ...\n"));
                return FALSE;
            }

            RtlInitializeBitMap( &BlockBitmap,
                                 BitmapCache,
                                 Length );

            dwBlk = RtlFindClearBits(&BlockBitmap, 1, 0);

            if (dwBlk != 0xFFFFFFFF)
            {
                break;
            }
            else
            {
                CcUnpinData(BitmapBcb);
                BitmapBcb = NULL;
                BitmapCache = NULL;

                RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
            }
        }
    }
        
    if (dwBlk < Length)
    {
        RtlSetBits(&BlockBitmap, dwBlk, 1);

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        *dwRet = dwBlk + EXT2_FIRST_DATA_BLOCK + Group * BLOCKS_PER_GROUP;

        //Updating Group Desc / Superblock
        Vcb->ext2_group_desc[Group].bg_free_blocks_count--;
        Ext2SaveGroup(IrpContext, Vcb);

        Vcb->ext2_super_block->s_free_blocks_count--;
        Ext2SaveSuper(IrpContext, Vcb);

        {
            ULONG i=0;
            for (i=0; i < Vcb->ext2_groups; i++)
            {
                if ((Vcb->ext2_group_desc[i].bg_block_bitmap == *dwRet) ||
                    (Vcb->ext2_group_desc[i].bg_inode_bitmap == *dwRet) ||
                    (Vcb->ext2_group_desc[i].bg_inode_table == *dwRet) )
                {
                    Ext2BreakPoint();
                    GroupHint = Group;
                    goto ScanBitmap;
                }
            }
        }

        return TRUE;
    }

    return FALSE;
}

BOOLEAN Ext2FreeBlock(  PEXT2_IRP_CONTEXT IrpContext,
                        PEXT2_VCB Vcb,
                        ULONG     Block )
{
    RTL_BITMAP      BlockBitmap;
    LARGE_INTEGER   Offset;
    ULONG           Length;

    PBCB            BitmapBcb;
    PVOID           BitmapCache;

    ULONG           Group, dwBlk;
    BOOLEAN         bModified = FALSE;

    if (Block < EXT2_FIRST_DATA_BLOCK || Block > (BLOCKS_PER_GROUP * Vcb->ext2_groups))
    {
        Ext2BreakPoint();
        return TRUE;
    }

    Ext2Print((DBG_INFO, "Ext2FreeBlock: Block %xh to be freed.\n", Block));

    Group = (Block - EXT2_FIRST_DATA_BLOCK) / BLOCKS_PER_GROUP;

    dwBlk = (Block - EXT2_FIRST_DATA_BLOCK) % BLOCKS_PER_GROUP;
    
    {
        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->ext2_group_desc[Group].bg_block_bitmap;

        if (Group == Vcb->ext2_groups - 1)
        {
            Length = TOTAL_BLOCKS % BLOCKS_PER_GROUP;

            /* s_blocks_count is integer multiple of s_blocks_per_group */
            if (Length == 0)
            {
                Length = BLOCKS_PER_GROUP;
            }
        }
        else
        {
            Length = BLOCKS_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) )
        {
            Ext2Print((DBG_ERROR, "Ext2DeleteBlock: PinReading error ...\n"));
            return FALSE;
        }

        RtlInitializeBitMap( &BlockBitmap,
                             BitmapCache,
                             Length );

        if (RtlCheckBit(&BlockBitmap, dwBlk) == 0)
        {
            
        }
        else
        {
            RtlClearBits(&BlockBitmap, dwBlk, 1);
            bModified = TRUE;
        }

        if (!bModified)
        {
            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&BlockBitmap, sizeof(RTL_BITMAP));
        }
    }
        
    if (bModified)
    {
        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        //Updating Group Desc / Superblock
        Vcb->ext2_group_desc[Group].bg_free_blocks_count++;
        Ext2SaveGroup(IrpContext, Vcb);

        Vcb->ext2_super_block->s_free_blocks_count++;
        Ext2SaveSuper(IrpContext, Vcb);

        return TRUE;
    }

    return FALSE;
}


BOOLEAN Ext2ExpandBlock(PEXT2_IRP_CONTEXT IrpContext,
                        PEXT2_VCB   Vcb,
                        PEXT2_FCB   Fcb,
                        ULONG   dwContent,
                        ULONG   Index,
                        ULONG   layer,
                        BOOLEAN bNew,
                        ULONG   *dwRet)
{
    ULONG       *pData = NULL;
    ULONG       i = 0, j = 0, temp = 1;
    ULONG       dwNewBlk = 0, dwBlk = 0;
    BOOLEAN     bDirty = FALSE;
    BOOLEAN     bRet = TRUE;

    PEXT2_INODE       Inode  = Fcb->ext2_inode;
    PEXT2_SUPER_BLOCK Ext2Sb = Vcb->ext2_super_block;

    pData = (ULONG *) ExAllocatePool(PagedPool, Vcb->BlockSize);

    if (!pData)
    {
        return FALSE;
    }

    RtlZeroMemory(pData, Vcb->BlockSize);

    if (bNew)
    {
        if (layer == 0)
        {
            if (IsDirectory(Fcb))
            {
                PEXT2_DIR_ENTRY2 pEntry;

                pEntry = (PEXT2_DIR_ENTRY2) pData;
                pEntry->rec_len = (USHORT)(Vcb->BlockSize);

                if (!Ext2SaveBlock(IrpContext, Vcb, dwContent, (PVOID)pData))
                {
                    bRet = FALSE;
                    goto errorout;
                }
            }
            else
            {
                LARGE_INTEGER   Offset;
            
                Offset.QuadPart  = (LONGLONG) dwContent;
                Offset.QuadPart = Offset.QuadPart * Vcb->BlockSize;

                Ext2RemoveMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);
            }
        }
        else
        {
            if (!Ext2SaveBlock(IrpContext, Vcb, dwContent, (PVOID)pData))
            {
                bRet = FALSE;
                goto errorout;
            }
        }
    }

    if (layer == 0)
    {
        dwNewBlk = dwContent;
    }
    else if (layer <= 3)
    {
        if (!bNew)
        {
            bRet = Ext2LoadBlock(Vcb, dwContent, (void *)pData);
            if (!bRet) goto errorout;
        }

        temp = 1 << ((BLOCK_BITS - 2) * (layer - 1));

        i = Index / temp;
        j = Index % temp;

        dwBlk = pData[i];

        if (dwBlk == 0)
        {
            if (!Ext2NewBlock(  IrpContext,
                                Vcb, 0,
                                dwContent,
                                &dwBlk)  )
            {
                bRet = FALSE;
                Ext2Print((DBG_ERROR, "Ext2ExpandBlock: get new block error.\n"));
                goto errorout;
            }

            Inode->i_blocks += (Vcb->BlockSize / SECTOR_SIZE);

            pData[i] = dwBlk;
            bDirty = TRUE;
        }

        if (!Ext2ExpandBlock( IrpContext,
                              Vcb, Fcb,
                              dwBlk, j,
                              layer - 1,
                              bDirty,
                              &dwNewBlk) )
        {
            bRet = FALSE;
            Ext2Print((DBG_ERROR, "Ext2ExpandBlockk: ... error recuise...\n"));
            goto errorout;
        }
        
        if (bDirty)
        {
            bRet = Ext2SaveBlock( IrpContext,
                                  Vcb, dwContent,
                                  (void *)pData );
        }
    }

errorout:

    if (pData)
        ExFreePool(pData);

    if (bRet && dwRet)
        *dwRet = dwNewBlk;

    return bRet;
}


BOOLEAN
Ext2ExpandInode( PEXT2_IRP_CONTEXT IrpContext,
                 PEXT2_VCB Vcb,
                 PEXT2_FCB Fcb,
                 ULONG *dwRet )
{
    ULONG dwSizes[EXT2_BLOCK_TYPES];
    ULONG Index = 0;
    ULONG dwTotal = 0;
    ULONG dwBlk = 0, dwNewBlk = 0;
    ULONG    i;
    BOOLEAN  bRet = FALSE;
    BOOLEAN  bNewBlock = FALSE;

    PEXT2_INODE Inode = Fcb->ext2_inode;

    Index = (ULONG)(Fcb->Header.AllocationSize.QuadPart >> BLOCK_BITS);

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        dwSizes[i] = Vcb->dwData[i];
        dwTotal += dwSizes[i];
    }

    if (Index >= dwTotal)
    {
        Ext2Print((DBG_ERROR, "Ext2ExpandInode: beyond the maxinum size of an inode.\n"));
        return FALSE;
    }

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        if (Index < dwSizes[i])
        {
            dwBlk = Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)];
            if (dwBlk == 0)
            {
                if (!Ext2NewBlock(IrpContext,
                            Vcb,
                            Fcb->BlkHint ? 0 : ((Fcb->Ext2Mcb->Inode - 1) / INODES_PER_GROUP),
                            Fcb->BlkHint,
                            &dwBlk ) )
                {
                    Ext2Print((DBG_ERROR, "Ext2ExpandInode: get new block error.\n"));
                    break;
                }

                Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)] = dwBlk;

                Inode->i_blocks += (Vcb->BlockSize / SECTOR_SIZE);

                bNewBlock = TRUE;
            }

            bRet = Ext2ExpandBlock( IrpContext,
                                    Vcb, Fcb,
                                    dwBlk, Index,
                                    i, bNewBlock,
                                    &dwNewBlk  ); 

            if (bRet)
            {
                Fcb->Header.AllocationSize.QuadPart += Vcb->BlockSize;
            }

            break;
        }

        Index -= dwSizes[i];
    }

    {
        ASSERT( Ext2DataBlocks(Vcb, Inode->i_blocks/(BLOCK_SIZE/SECTOR_SIZE))
                == (Fcb->Header.AllocationSize.LowPart / BLOCK_SIZE));

        ASSERT( Ext2TotalBlocks(Vcb, Fcb->Header.AllocationSize.LowPart/BLOCK_SIZE)
                == (Inode->i_blocks/(BLOCK_SIZE/SECTOR_SIZE)));
    }


    Ext2SaveInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, Inode);

    if (bRet && dwNewBlk)
    {
        if (dwRet)
        {
            Fcb->BlkHint = dwNewBlk+1;
            *dwRet = dwNewBlk;

            Ext2Print((DBG_INFO, "Ext2ExpandInode: %S (%xh) i=%2.2xh Index=%8.8xh New Block=%8.8xh\n", Fcb->Ext2Mcb->ShortName.Buffer, Fcb->Ext2Mcb->Inode, i, Index, dwNewBlk));
        }

        return TRUE;
    }

    return FALSE;
}

NTSTATUS
Ext2NewInode(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB Vcb,
            ULONG   GroupHint,
            ULONG   Type,
            PULONG  Inode )
{
    RTL_BITMAP      InodeBitmap;
    PVOID           BitmapCache;
    PBCB            BitmapBcb;

    ULONG           Group, i, j;
    ULONG           Average, Length;
    LARGE_INTEGER   Offset;
    
    ULONG           dwInode;

    *Inode = dwInode = 0XFFFFFFFF;

repeat:

    Group = i = 0;
    
    if (Type == EXT2_FT_DIR)
    {
        Average = Vcb->ext2_super_block->s_free_inodes_count / Vcb->ext2_groups;

        for (j = 0; j < Vcb->ext2_groups; j++)
        {
            i = (j + GroupHint) % (Vcb->ext2_groups);

            if ((Vcb->ext2_group_desc[i].bg_used_dirs_count << 8) < 
                 Vcb->ext2_group_desc[i].bg_free_inodes_count )
            {
                Group = i + 1;
                break;
            }
        }

        if (!Group)
        {
            for (j = 0; j < Vcb->ext2_groups; j++)
            {
                if (Vcb->ext2_group_desc[j].bg_free_inodes_count >= Average)
                {
                    if (!Group || (Vcb->ext2_group_desc[j].bg_free_blocks_count > Vcb->ext2_group_desc[Group].bg_free_blocks_count ))
                        Group = j + 1;
                }
            }
        }
    }
    else 
    {
        /*
         * Try to place the inode in its parent directory (GroupHint)
         */
        if (Vcb->ext2_group_desc[GroupHint].bg_free_inodes_count)
        {
            Group = GroupHint + 1;
        }
        else
        {
            i = GroupHint;

            /*
             * Use a quadratic hash to find a group with a
             * free inode
             */
            for (j = 1; j < Vcb->ext2_groups; j <<= 1)
            {

                i += j;
                if (i > Vcb->ext2_groups) 
                    i -= Vcb->ext2_groups;

                if (Vcb->ext2_group_desc[i].bg_free_inodes_count)
                {
                    Group = i + 1;
                    break;
                }
            }
        }

        if (!Group) {
            /*
             * That failed: try linear search for a free inode
             */
            i = GroupHint + 1;
            for (j = 2; j < Vcb->ext2_groups; j++)
            {
                if (++i >= Vcb->ext2_groups) i = 0;

                if (Vcb->ext2_group_desc[i].bg_free_inodes_count)
                {
                    Group = i + 1;
                    break;
                }
            }
        }
    }

    // Could not find a proper group.
    if (!Group)
    {
        return STATUS_DISK_FULL;
    }
    else
    {
        Group--;

        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->ext2_group_desc[Group].bg_inode_bitmap;

        if (Vcb->ext2_groups == 1)
        {
            Length = INODES_COUNT;
        }
        else
        {
            if (Group == Vcb->ext2_groups - 1)
            {
                Length = INODES_COUNT % INODES_PER_GROUP;
                if (!Length) 
                {
                    /* INODES_COUNT is integer multiple of INODES_PER_GROUP */
                    Length = INODES_PER_GROUP;
                }
            }
            else
            {
                Length = INODES_PER_GROUP;
            }
        }
        
        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) )
        {
            Ext2Print((DBG_ERROR, "Ext2NewInode: PinReading error ...\n"));

            return STATUS_UNSUCCESSFUL;
        }

        RtlInitializeBitMap( &InodeBitmap,
                             BitmapCache,
                             Length );

        dwInode = RtlFindClearBits(&InodeBitmap, 1, 0);

        if (dwInode == 0xFFFFFFFF)
        {
            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&InodeBitmap, sizeof(RTL_BITMAP));
        }
    }

    if (dwInode == 0xFFFFFFFF || dwInode >= Length)
    {
        if (Vcb->ext2_group_desc[Group].bg_free_inodes_count != 0)
        {
            Vcb->ext2_group_desc[Group].bg_free_inodes_count = 0;

            Ext2SaveGroup(IrpContext, Vcb);            
        }

        goto repeat;
    }
    else
    {
        RtlSetBits(&InodeBitmap, dwInode, 1);

        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        *Inode = dwInode + 1 + Group * INODES_PER_GROUP;

        //Updating Group Desc / Superblock
        Vcb->ext2_group_desc[Group].bg_free_inodes_count--;
        if (Type == EXT2_FT_DIR)
        {
            Vcb->ext2_group_desc[Group].bg_used_dirs_count++;
        }

        Ext2SaveGroup(IrpContext, Vcb);

        Vcb->ext2_super_block->s_free_inodes_count--;
        Ext2SaveSuper(IrpContext, Vcb);
        
        return STATUS_SUCCESS;        
    }

    return STATUS_DISK_FULL;
}

BOOLEAN
Ext2FreeInode(
            PEXT2_IRP_CONTEXT IrpContext,
            PEXT2_VCB Vcb,
            ULONG Inode,
            ULONG Type )
{
    RTL_BITMAP      InodeBitmap;
    PVOID           BitmapCache;
    PBCB            BitmapBcb;

    ULONG           Group;
    ULONG           Length;
    LARGE_INTEGER   Offset;

    ULONG           dwIno;
    BOOLEAN         bModified = FALSE;
    

    Group = (Inode - 1) / INODES_PER_GROUP;
    dwIno = (Inode - 1) % INODES_PER_GROUP;

    Ext2Print((DBG_INFO, "Ext2FreeInode: Inode: %xh (Group/Off = %xh/%xh)\n",
                         Inode, Group, dwIno));
    
    {
        Offset.QuadPart = (LONGLONG) Vcb->BlockSize;
        Offset.QuadPart = Offset.QuadPart * Vcb->ext2_group_desc[Group].bg_inode_bitmap;
        if (Group == Vcb->ext2_groups - 1)
        {
            Length = INODES_COUNT % INODES_PER_GROUP;
            if (!Length)
            { /* s_inodes_count is integer multiple of s_inodes_per_group */
                Length = INODES_PER_GROUP;
            }
        }
        else
        {
            Length = INODES_PER_GROUP;
        }

        if (!CcPinRead( Vcb->StreamObj,
                        &Offset,
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &BitmapBcb,
                        &BitmapCache ) )
        {
            Ext2Print((DBG_ERROR, "Ext2FreeInode: PinReading error ...\n"));
            return FALSE;
        }

        RtlInitializeBitMap( &InodeBitmap,
                             BitmapCache,
                             Length );

        if (RtlCheckBit(&InodeBitmap, dwIno) == 0)
        {
            
        }
        else
        {
            RtlClearBits(&InodeBitmap, dwIno, 1);
            bModified = TRUE;
        }

        if (!bModified)
        {
            CcUnpinData(BitmapBcb);
            BitmapBcb = NULL;
            BitmapCache = NULL;

            RtlZeroMemory(&InodeBitmap, sizeof(RTL_BITMAP));
        }
    }
        
    if (bModified)
    {
        CcSetDirtyPinnedData(BitmapBcb, NULL );

        Ext2RepinBcb(IrpContext, BitmapBcb);

        CcUnpinData(BitmapBcb);

        Ext2AddMcbEntry(Vcb, Offset.QuadPart, (LONGLONG)Vcb->BlockSize);

        //Updating Group Desc / Superblock
        if (Type == EXT2_FT_DIR)
        {
            Vcb->ext2_group_desc[Group].bg_used_dirs_count--;
        }

        Vcb->ext2_group_desc[Group].bg_free_inodes_count++;
        Ext2SaveGroup(IrpContext, Vcb);

        Vcb->ext2_super_block->s_free_inodes_count++;
        Ext2SaveSuper(IrpContext, Vcb);

        return TRUE;
    }

    return FALSE;
}


NTSTATUS
Ext2AddEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               FileType,
         IN ULONG               Inode,
         IN PUNICODE_STRING     FileName )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pDir = NULL;
    PEXT2_DIR_ENTRY2        pNewDir = NULL;
    PEXT2_DIR_ENTRY2        pTarget = NULL;

    ULONG                   Length = 0;
    ULONG                   dwBytes = 0;

    BOOLEAN                 bFound = FALSE;
    BOOLEAN                 bAdding = FALSE;

    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   dwRet;

    if (!IsDirectory(Dcb))
    {
        Ext2BreakPoint();
        Status = STATUS_INVALID_PARAMETER;
        return Status;
    }

    MainResourceAcquired = ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try
    {
        Dcb->ReferenceCount++;

        pDir = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                    EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
        if (!pDir)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        pTarget = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                     2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
        if (!pTarget)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        if (IsFlagOn( SUPER_BLOCK->s_feature_incompat, 
                      EXT2_FEATURE_INCOMPAT_FILETYPE))
        {
            pDir->file_type = (UCHAR) FileType;
        }
        else
        {
            pDir->file_type = 0;
        }

        {
            OEM_STRING  OemName;
            OemName.Buffer = pDir->name;
            OemName.MaximumLength = EXT2_NAME_LEN;
            OemName.Length  = 0;

            Status = Ext2UnicodeToOEM(&OemName, FileName);

            if (!NT_SUCCESS(Status))
            {
                __leave;
            }

            pDir->name_len = (CCHAR) OemName.Length;
        }

        pDir->inode  = Inode;
        pDir->rec_len = (USHORT) (EXT2_DIR_REC_LEN(pDir->name_len));

        dwBytes = 0;

Repeat:

        while ((LONGLONG)dwBytes < Dcb->Header.AllocationSize.QuadPart)
        {
            RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

            // Reading the DCB contents
            Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->ext2_inode,
                        dwBytes,
                        (PVOID)pTarget,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        &dwRet);

            if (!NT_SUCCESS(Status))
            {
                Ext2Print((DBG_ERROR, "Ext2AddDirectory: Reading Directory Content error.\n"));
                __leave;
            }

            if (((pTarget->inode == 0) && pTarget->rec_len >= pDir->rec_len) || 
                (pTarget->rec_len >= EXT2_DIR_REC_LEN(pTarget->name_len) + pDir->rec_len))
            {
                if (pTarget->inode)
                {
                    RtlZeroMemory(pTarget, 2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

                    // Reading the DCB contents
                     Status = Ext2ReadInode(
                                NULL,
                                Vcb,
                                Dcb->ext2_inode,
                                dwBytes,
                                (PVOID)pTarget,
                                2 * EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                                &dwRet);

                    if (!NT_SUCCESS(Status))
                    {
                        Ext2Print((DBG_ERROR, "Ext2AddDirectory: Reading Directory Content error.\n"));
                        __leave;
                    }

                    Length = EXT2_DIR_REC_LEN(pTarget->name_len);
                    
                    pNewDir = (PEXT2_DIR_ENTRY2) ((PUCHAR)pTarget + EXT2_DIR_REC_LEN(pTarget->name_len));

                    pNewDir->rec_len = pTarget->rec_len - EXT2_DIR_REC_LEN(pTarget->name_len);

                    pTarget->rec_len = EXT2_DIR_REC_LEN(pTarget->name_len);
                }
                else
                {
                    pNewDir = pTarget;
                    pNewDir->rec_len = (USHORT)((ULONG)(Dcb->Header.AllocationSize.QuadPart) - dwBytes);
                }

                pNewDir->file_type = pDir->file_type;
                pNewDir->inode = pDir->inode;
                pNewDir->name_len = pDir->name_len;
                memcpy(pNewDir->name, pDir->name, pDir->name_len);
                Length += EXT2_DIR_REC_LEN(pDir->name_len);

                bFound = TRUE;
                break;
            }
            
            dwBytes += pTarget->rec_len;
        }

        if (bFound)
        {
            ULONG   dwRet;

            if ( FileType==EXT2_FT_DIR )
            {
                if(((pDir->name_len == 1) && (pDir->name[0] == '.')) ||
                   ((pDir->name_len == 2) && (pDir->name[0] == '.') && (pDir->name[1] == '.')) )
                {
                }
                else
                {
                    Dcb->ext2_inode->i_links_count++;
                }
            }

            Status = Ext2WriteInode(IrpContext, Vcb, Dcb->ext2_inode, dwBytes, pTarget, Length, FALSE, &dwRet);
        }
        else
        {
            // We should expand the size of the dir inode 
            if (!bAdding)
            {
                ULONG dwRet;

                bAdding = Ext2ExpandInode(IrpContext, Vcb, Dcb, &dwRet);

                if (bAdding) {

                    Dcb->ext2_inode->i_size = Dcb->Header.AllocationSize.LowPart;

                    Ext2SaveInode(IrpContext, Vcb, Dcb->Ext2Mcb->Inode, Dcb->ext2_inode);

                    Dcb->Header.FileSize = Dcb->Header.AllocationSize;

                    goto Repeat;
                }

                __leave;

            }
            else  // Something must be error!
            {
                __leave;
            }
        }
    }

    __finally
    {

        Dcb->ReferenceCount--;

        if(MainResourceAcquired)    {
            ExReleaseResourceForThreadLite(
                    &Dcb->MainResource,
                    ExGetCurrentResourceThread());
        }

        if (pTarget != NULL)
        {
            ExFreePool(pTarget);
        }

        if (pDir)   {
            ExFreePool(pDir);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2RemoveEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               FileType,
         IN ULONG               Inode )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pTarget = NULL;
    PEXT2_DIR_ENTRY2        pPrevDir = NULL;

    USHORT                  PrevRecLen = 0;

    ULONG                   Length = 0;
    ULONG                   dwBytes = 0;

    BOOLEAN                 bRet = FALSE;
    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   dwRet;

    if (!IsDirectory(Dcb))
    {
        return FALSE;
    }

    MainResourceAcquired = 
            ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try
    {

        Dcb->ReferenceCount++;

        pTarget = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                     EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
        if (!pTarget)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }
        
        pPrevDir = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                     EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
        if (!pPrevDir)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        dwBytes = 0;

        while ((LONGLONG)dwBytes < Dcb->Header.AllocationSize.QuadPart)
        {
            RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));

            Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->ext2_inode,
                        dwBytes,
                        (PVOID)pTarget,
                        EXT2_DIR_REC_LEN(EXT2_NAME_LEN),
                        &dwRet);

            if (!NT_SUCCESS(Status))
            {
                Ext2Print((DBG_ERROR, "Ext2RemoveEntry: Reading Directory Content error.\n"));
                __leave;
            }

            if (pTarget->inode == Inode)
            {
                ULONG   dwRet;
                ULONG   RecLen;

                if ((PrevRecLen + pTarget->rec_len) < MAXIMUM_RECORD_LENGTH )
                {
                    pPrevDir->rec_len += pTarget->rec_len;
                    RecLen = EXT2_DIR_REC_LEN(pTarget->name_len);

                    RtlZeroMemory(pTarget, RecLen);

                    Ext2WriteInode(IrpContext, Vcb, Dcb->ext2_inode, dwBytes - PrevRecLen, pPrevDir, 8, FALSE, &dwRet);
                    Ext2WriteInode(IrpContext, Vcb, Dcb->ext2_inode, dwBytes, pTarget, RecLen, FALSE, &dwRet);
                }
                else
                {
                    RecLen = (ULONG)pTarget->rec_len;
                    if (RecLen > EXT2_DIR_REC_LEN(EXT2_NAME_LEN))
                    {
                        RtlZeroMemory(pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
                    }
                    else
                    {
                        RtlZeroMemory(pTarget, RecLen);
                    }

                    pTarget->rec_len = (USHORT)RecLen;

                    Ext2WriteInode(IrpContext, Vcb, Dcb->ext2_inode, dwBytes, pTarget, RecLen, FALSE, &dwRet);
                }

                if ( FileType==EXT2_FT_DIR )
                {
                    if(((pTarget->name_len == 1) && (pTarget->name[0] == '.')) ||
                       ((pTarget->name_len == 2) && (pTarget->name[0] == '.') && (pTarget->name[1] == '.')) )
                    {
                        Ext2BreakPoint();
                    }
                    else
                    {
                        Dcb->ext2_inode->i_links_count--;
                    }
                }

                /* Update at least mtime/atime if !EXT2_FT_DIR. */
                Ext2SaveInode(IrpContext, Vcb, Dcb->Ext2Mcb->Inode, Dcb->ext2_inode);

                bRet = TRUE;
                
                break;
            }
            else
            {
                RtlCopyMemory(pPrevDir, pTarget, EXT2_DIR_REC_LEN(EXT2_NAME_LEN));
                PrevRecLen = pTarget->rec_len;
            }

            dwBytes += pTarget->rec_len;
        }
    }

    __finally
    {

        Dcb->ReferenceCount--;

        if(MainResourceAcquired)
            ExReleaseResourceForThreadLite(
                    &Dcb->MainResource,
                    ExGetCurrentResourceThread());

        if (pTarget != NULL)
        {
            ExFreePool(pTarget);
        }

        if (pPrevDir != NULL)
        {
            ExFreePool(pPrevDir);
        }
    }
    
    return bRet;
}

NTSTATUS
Ext2SetParentEntry (
         IN PEXT2_IRP_CONTEXT   IrpContext,
         IN PEXT2_VCB           Vcb,
         IN PEXT2_FCB           Dcb,
         IN ULONG               OldParent,
         IN ULONG               NewParent )
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;

    PEXT2_DIR_ENTRY2        pSelf   = NULL;
    PEXT2_DIR_ENTRY2        pParent = NULL;

    ULONG                   dwBytes = 0;

    BOOLEAN                 MainResourceAcquired = FALSE;

    ULONG                   Offset = 0;

    if (!IsDirectory(Dcb))
    {
        Status = STATUS_INVALID_PARAMETER;
        return Status;
    }

    MainResourceAcquired = 
        ExAcquireResourceExclusiveLite(&Dcb->MainResource, TRUE);

    __try
    {
        Dcb->ReferenceCount++;

        pSelf = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                   EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2));
        if (!pSelf)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        dwBytes = 0;

        // Reading the DCB contents
        Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        Dcb->ext2_inode,
                        Offset,
                        (PVOID)pSelf,
                        EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2),
                        &dwBytes);

        if (!NT_SUCCESS(Status))
        {
            Ext2Print((DBG_ERROR, "Ext2SetParentEntry: Reading Directory Content error.\n"));
            __leave;
        }

        ASSERT(dwBytes == EXT2_DIR_REC_LEN(1) + EXT2_DIR_REC_LEN(2));

        pParent = (PEXT2_DIR_ENTRY2)((PUCHAR)pSelf + pSelf->rec_len);

        if (pParent->inode != OldParent)
        {
            Ext2BreakPoint();
        }

        pParent->inode = NewParent;

        Status = Ext2WriteInode( IrpContext,
                                 Vcb, 
                                 Dcb->ext2_inode,
                                 Offset,
                                 pSelf,
                                 dwBytes,
                                 FALSE,
                                 &dwBytes );
    }

    __finally
    {
        Dcb->ReferenceCount--;

        if(MainResourceAcquired)    {
            ExReleaseResourceForThreadLite(
                    &Dcb->MainResource,
                    ExGetCurrentResourceThread());
        }

        if (pSelf)
        {
            ExFreePool(pSelf);
        }
    }
    
    return Status;
}


BOOLEAN Ext2TruncateBlock(
         IN PEXT2_IRP_CONTEXT IrpContext,
         IN PEXT2_VCB Vcb,
         IN PEXT2_FCB Fcb,
         IN ULONG   dwContent,
         IN ULONG   Index,
         IN ULONG   layer,
         OUT BOOLEAN *bFreed )
{
    ULONG       *pData = NULL;
    ULONG       i = 0, j = 0, temp = 1;
    BOOLEAN     bDirty = FALSE;
    BOOLEAN     bRet = FALSE;
    ULONG       dwBlk;

    LONGLONG    Offset;

    PBCB        Bcb;

    PEXT2_INODE Inode = Fcb->ext2_inode;

    *bFreed = FALSE;

    if (layer == 0)
    {
        if (dwContent > 0 && dwContent < (BLOCKS_PER_GROUP * Vcb->ext2_groups))
        {
            bRet = Ext2FreeBlock(IrpContext, Vcb, dwContent);

            if (bRet)
            {
                ASSERT(Inode->i_blocks >= (Vcb->BlockSize / SECTOR_SIZE));
                Inode->i_blocks -= (Vcb->BlockSize / SECTOR_SIZE);            
            }
        }
        else
        {
            Ext2BreakPoint();
            bRet = FALSE;
        }

        *bFreed = bRet;
    }
    else if (layer <= 3)
    {
        Offset = (LONGLONG) dwContent;
        Offset = Offset * Vcb->BlockSize;

        if( !CcPinRead( Vcb->StreamObj,
                        (PLARGE_INTEGER) (&Offset),
                        Vcb->BlockSize,
                        PIN_WAIT,
                        &Bcb,
                        &pData ))
        {
            Ext2Print((DBG_ERROR, "Ext2SaveBuffer: PinReading error ...\n"));
            goto errorout;
        }

        temp = 1 << ((BLOCK_BITS - 2) * (layer - 1));

        i = Index / temp;
        j = Index % temp;

        dwBlk = pData[i];

        if(!Ext2TruncateBlock(IrpContext, Vcb, Fcb, dwBlk, j, layer - 1, &bDirty))
        {
            goto errorout;
        }

        if (bDirty)
        {
            pData[i] = 0;
        }

        if (i == 0 && j == 0)
        {
            CcUnpinData(Bcb);
            pData = NULL;

            *bFreed = TRUE;
            bRet = Ext2FreeBlock(IrpContext, Vcb, dwContent);

            if (bRet)
            {
                ASSERT(Inode->i_blocks >= (Vcb->BlockSize / SECTOR_SIZE));
                Inode->i_blocks -= (Vcb->BlockSize / SECTOR_SIZE);
            }
        }
        else
        {
            CcSetDirtyPinnedData(Bcb, NULL );
            Ext2RepinBcb(IrpContext, Bcb);

            Ext2AddMcbEntry(Vcb, Offset, (LONGLONG)Vcb->BlockSize);

            bRet = TRUE;
            *bFreed = FALSE;
        }
    }

errorout:

    if (pData)
    {
        CcUnpinData(Bcb);
    }

    return bRet;
}


BOOLEAN
Ext2TruncateInode(
         IN PEXT2_IRP_CONTEXT IrpContext,
         IN PEXT2_VCB   Vcb,
         IN PEXT2_FCB   Fcb )
{
    ULONG   dwSizes[EXT2_BLOCK_TYPES];
    ULONG   Index = 0;
    ULONG   dwTotal = 0;
    ULONG   dwBlk = 0;

    ULONG    i;
    BOOLEAN  bRet = FALSE;
    BOOLEAN  bFreed = FALSE;

    PEXT2_INODE Inode = Fcb->ext2_inode;

    Index = (ULONG)(Fcb->Header.AllocationSize.QuadPart >> BLOCK_BITS);

    if (Index > 0) 
    {
        Index--;
    }
    else
    {
        return TRUE;
    }

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        dwSizes[i] = Vcb->dwData[i];
        dwTotal += dwSizes[i];
    }

    if (Index >= dwTotal)
    {
        Ext2Print((DBG_ERROR, "Ext2ExpandInode: beyond the maxinum size of an inode.\n"));
        return TRUE;
    }

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        if (Index < dwSizes[i])
        {
            dwBlk = Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)];

            bRet = Ext2TruncateBlock(IrpContext, Vcb, Fcb, dwBlk, Index , i, &bFreed); 

            if (bRet)
            {
                Fcb->Header.AllocationSize.QuadPart -= Vcb->BlockSize;
            
                if (bFreed)
                {
                    Inode->i_block[i==0 ? (Index):(i + EXT2_NDIR_BLOCKS - 1)] = 0;
                }
            }

            break;
        }

        Index -= dwSizes[i];
    }

    {
        ASSERT( Ext2DataBlocks(Vcb, Inode->i_blocks/(BLOCK_SIZE/SECTOR_SIZE))
                == (Fcb->Header.AllocationSize.LowPart / BLOCK_SIZE));

        ASSERT( Ext2TotalBlocks(Vcb, Fcb->Header.AllocationSize.LowPart/BLOCK_SIZE)
                == (Inode->i_blocks/(BLOCK_SIZE/SECTOR_SIZE)));

    }

    //
    // Inode struct saving is done externally.
    //

    Ext2SaveInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, Inode);


    return bRet;
}

BOOLEAN
Ext2AddMcbEntry (
    IN PEXT2_VCB Vcb,
    IN LONGLONG  Lba,
    IN LONGLONG  Length)
{
    BOOLEAN     bRet = FALSE;

    LONGLONG    Offset;

#if DBG
    LONGLONG    DirtyLba;
    LONGLONG    DirtyLen;
#endif


    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));

    Length = (Length + Lba - Offset + BLOCK_SIZE - 1) &
             (~((LONGLONG)BLOCK_SIZE - 1));

    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);
    ASSERT ((Length & (BLOCK_SIZE - 1)) == 0);

    Offset = (Offset >> BLOCK_BITS) + 1;
    Length = (Length >>BLOCK_BITS);

    ExAcquireResourceExclusiveLite(
        &(Vcb->McbResource),
        TRUE );

    Ext2Print((DBG_INFO, "Ext2AddMcbEntry: Lba=%I64xh Length=%I64xh\n",
                          Offset, Length ));

#if DBG
    bRet = FsRtlLookupLargeMcbEntry(
                    &(Vcb->DirtyMcbs),
                    Offset,
                    &DirtyLba,
                    &DirtyLen,
                    NULL,
                    NULL,
                    NULL );

    if (bRet && DirtyLba == Offset && DirtyLen >= Length)
    {
        Ext2Print((DBG_INFO, "Ext2AddMcbEntry: this run already exists.\n"));
    }
#endif

    __try
    {
        bRet = FsRtlAddLargeMcbEntry(
                         &(Vcb->DirtyMcbs),
                         Offset, Offset,
                         Length );

    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        Ext2BreakPoint();
        bRet = FALSE;
    }

#if DBG
    if (bRet)
    {
        BOOLEAN     bFound = FALSE;
        LONGLONG    RunStart;
        LONGLONG    RunLength;
        ULONG       Index;

        bFound = FsRtlLookupLargeMcbEntry(
                        &(Vcb->DirtyMcbs),
                        Offset,
                        &DirtyLba,
                        &DirtyLen,
                        &RunStart,
                        &RunLength,
                        &Index );

        if ((!bFound) || (DirtyLba == -1) ||
            (DirtyLba != Offset) || (DirtyLen < Length))
        {
            LONGLONG            DirtyVba;
            LONGLONG            DirtyLba;
            LONGLONG            DirtyLength;

            Ext2BreakPoint();

            for ( Index = 0; 
                  FsRtlGetNextLargeMcbEntry( &(Vcb->DirtyMcbs),
                                             Index,
                                             &DirtyVba,
                                             &DirtyLba,
                                             &DirtyLength); 
                  Index++ )
            {
                Ext2Print((DBG_INFO, "Index = %xh\n", Index));
                Ext2Print((DBG_INFO, "DirtyVba = %I64xh\n", DirtyVba));
                Ext2Print((DBG_INFO, "DirtyLba = %I64xh\n", DirtyLba));
                Ext2Print((DBG_INFO, "DirtyLen = %I64xh\n\n", DirtyLength));
            }
        }
    }
#endif

    ExReleaseResourceForThreadLite(
        &(Vcb->McbResource),
        ExGetCurrentResourceThread() );

    return bRet;
}

VOID
Ext2RemoveMcbEntry (
    IN PEXT2_VCB Vcb,
    IN LONGLONG  Lba,
    IN LONGLONG  Length)
{
    LONGLONG Offset;

    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));

    Length = (Length + Lba - Offset + BLOCK_SIZE - 1) &
             (~((LONGLONG)BLOCK_SIZE - 1));

    ASSERT(Offset == Lba);

    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);
    ASSERT ((Length & (BLOCK_SIZE - 1)) == 0);

    Offset = (Offset >> BLOCK_BITS) + 1;
    Length = (Length >> BLOCK_BITS);

    Ext2Print((DBG_INFO, "Ext2RemoveMcbEntry: Lba=%I64xh Length=%I64xh\n",
                          Offset, Length ));

    ExAcquireResourceExclusiveLite(
        &(Vcb->McbResource),
        TRUE );

    __try
    {
        FsRtlRemoveLargeMcbEntry(
            &(Vcb->DirtyMcbs),
            Offset, Length      );

    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        Ext2BreakPoint();
    }

#if DBG
    {
        BOOLEAN  bFound = FALSE;
        LONGLONG DirtyLba, DirtyLen;

        bFound = FsRtlLookupLargeMcbEntry(
                        &(Vcb->DirtyMcbs),
                        Offset,
                        &DirtyLba,
                        &DirtyLen,
                        NULL,
                        NULL,
                        NULL );

        if (bFound &&( DirtyLba != -1))
        {
            Ext2BreakPoint();
        }
    }
#endif

    ExReleaseResourceForThreadLite(
        &(Vcb->McbResource),
        ExGetCurrentResourceThread() );
}

BOOLEAN
Ext2LookupMcbEntry (
    IN PEXT2_VCB    Vcb,
    IN LONGLONG     Lba,
    OUT PLONGLONG   pLba,
    OUT PLONGLONG   pLength,
    OUT PLONGLONG   RunStart,
    OUT PLONGLONG   RunLength,
    OUT PULONG      Index)
{
    BOOLEAN     bRet;
    LONGLONG    Offset;


    Offset = Lba & (~((LONGLONG)BLOCK_SIZE - 1));
    ASSERT ((Offset & (BLOCK_SIZE - 1)) == 0);

    ASSERT(Lba == Offset);

    Offset = (Offset >> BLOCK_BITS) + 1;

    ExAcquireResourceExclusiveLite(
        &(Vcb->McbResource),
        TRUE );

    bRet = FsRtlLookupLargeMcbEntry(
                &(Vcb->DirtyMcbs),
                Offset,
                pLba,
                pLength,
                RunStart,
                RunLength,
                Index );

    ExReleaseResourceForThreadLite(
        &(Vcb->McbResource),
        ExGetCurrentResourceThread() );

    if (bRet)
    {
        if (pLba && ((*pLba) != -1))
        {
            ASSERT((*pLba) > 0);

            (*pLba) = (((*pLba) - 1) << BLOCK_BITS);
            (*pLba) += ((Lba) & ((LONGLONG)BLOCK_SIZE - 1));
        }

        if (pLength)
        {
            (*pLength) <<= BLOCK_BITS;
            (*pLength)  -= ((Lba) & ((LONGLONG)BLOCK_SIZE - 1));
        }

        if (RunStart && (*RunStart != -1))
        {
            (*RunStart) = (((*RunStart) - 1) << BLOCK_BITS);
        }

        if (RunLength)
        {
            (*RunLength) <<= BLOCK_BITS;
        }
    }

    return bRet;
}


ULONG
Ext2DataBlocks(PEXT2_VCB Vcb, ULONG TotalBlocks)
{
    ULONG   dwData[EXT2_BLOCK_TYPES];
    ULONG   dwMeta[EXT2_BLOCK_TYPES];
    ULONG   DataBlocks = 0;
    ULONG   i, j;

    if (TotalBlocks <= EXT2_NDIR_BLOCKS)
    {
        return TotalBlocks;
    }

    TotalBlocks -= EXT2_NDIR_BLOCKS;

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        if (i == 0)
        {
            dwData[i] = 1;
        }
        else
        {
            dwData[i] = Vcb->dwData[i];
        }

        dwMeta[i] = Vcb->dwMeta[i];
    }

    for( i=1; (i < EXT2_BLOCK_TYPES) && (TotalBlocks > 0); i++)
    {
        if (TotalBlocks >= (dwData[i] + dwMeta[i]))
        {
            TotalBlocks -= (dwData[i] + dwMeta[i]);
            DataBlocks  += dwData[i];
        }
        else
        {
            ULONG   dwDivide = 0;
            ULONG   dwRemain = 0;

            for (j=i; (j > 0) && (TotalBlocks > 0); j--)
            {
                dwDivide = (TotalBlocks - 1) / (dwData[j-1] + dwMeta[j-1]);
                dwRemain = (TotalBlocks - 1) % (dwData[j-1] + dwMeta[j-1]);

                DataBlocks += (dwDivide * dwData[j-1]);
                TotalBlocks = dwRemain;
            }
        }
    }

    return (DataBlocks + EXT2_NDIR_BLOCKS);
}


ULONG
Ext2TotalBlocks(PEXT2_VCB Vcb, ULONG DataBlocks)
{
    ULONG   dwData[EXT2_BLOCK_TYPES];
    ULONG   dwMeta[EXT2_BLOCK_TYPES];
    ULONG   TotalBlocks = 0;
    ULONG   i, j;

    if (DataBlocks <= EXT2_NDIR_BLOCKS)
    {
        return DataBlocks;
    }

    DataBlocks -= EXT2_NDIR_BLOCKS;

    for (i = 0; i < EXT2_BLOCK_TYPES; i++)
    {
        if (i == 0)
        {
            dwData[i] = 1;
        }
        else
        {
            dwData[i] = Vcb->dwData[i];
        }

        dwMeta[i] = Vcb->dwMeta[i];
    }

    for( i=1; (i < EXT2_BLOCK_TYPES) && (DataBlocks > 0); i++)
    {
        if (DataBlocks >= dwData[i])
        {
            DataBlocks  -= dwData[i];
            TotalBlocks += (dwData[i] + dwMeta[i]);
        }
        else
        {
            ULONG   dwDivide = 0;
            ULONG   dwRemain = 0;

            for (j=i; (j > 0) && (DataBlocks > 0); j--)
            {
                dwDivide = (DataBlocks) / (dwData[j-1]);
                dwRemain = (DataBlocks) % (dwData[j-1]);

                TotalBlocks += (dwDivide * (dwData[j-1] + dwMeta[j-1]) + 1);
                DataBlocks = dwRemain;
            }
        }
    }

    return (TotalBlocks + EXT2_NDIR_BLOCKS);
}
